"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/util.py $
$Id: util.py 24750 2004-07-21 15:26:51Z dbinger $
"""

import sys
import __builtin__
from types import TypeType, ClassType, InstanceType


def is_class_object (klass):
    """Return true if 'klass' is a class object (either a regular Python
    class or an ExtensionClass).
    """
    # The first clause handles regular classes quickly; the second
    # is needed to accomodate ExtensionClasses
    return type(klass) is ClassType or hasattr(klass, '__bases__')

def is_instance_object (object):
    """Return true if 'object' is an instance of some class (either a
    regular Python class or an ExtensionClass).
    """
    return ((type(object) is InstanceType or
             (hasattr(object, '__class__') and
              not hasattr(object, '__bases__'))) and
            hasattr(object, '__dict__'))

def get_class_object (name):
    lastdot = name.rfind(".")
    if lastdot == -1:                   # no dots -- look in main module
        module_name = '__main__'
        klass_name = name
        module = sys.modules[module_name]
    else:
        module_name = name[:lastdot]
        klass_name = name[lastdot+1:]
        try:
            __import__(module_name)
        except ImportError, exc:
            raise ValueError, \
                  "error finding class %s: %s" % (name, exc)
        module = sys.modules[module_name]

    if not hasattr(module, klass_name):
        raise ValueError, \
              "no such class %s in module %s" % (`klass_name`, module_name)
    else:
        klass = getattr(module, klass_name)
        if is_class_object(klass):
            return klass
        else:
            raise ValueError, \
                  "%s in module %s: not a class" % (`klass_name`, module_name)


def get_full_classname (klass):
    classname = klass.__name__
    modname = klass.__module__
    if modname == "__main__":
        return classname
    else:
        return "%s.%s" % (modname, classname)


# Supposedly, these two won't be necessary with Python 2.1, since its
# 'issubclass()' and 'isinstance()' will finally be able to deal with
# extension classes completely in 2.1.  (It's close in 1.6 and 2.0, but not
# quite there yet.)  I have not yet verified this!

def issubclass (class1, class2):
    """A version of 'issubclass()' that works with extension classes
    as well as regular Python classes.
    """

    # Both class objects are regular Python classes, so use the
    # built-in 'issubclass()'.
    if type(class1) is ClassType and type(class2) is ClassType:
        return __builtin__.issubclass(class1, class2)

    # Both so-called class objects have a '__bases__' attribute: ie.,
    # they aren't regular Python classes, but they sure look like them.
    # Assume they are extension classes and reimplement what the builtin
    # 'issubclass()' does behind the scenes.
    elif hasattr(class1, '__bases__') and hasattr(class2, '__bases__'):
        # XXX it appears that "ec.__class__ is type(ec)" for an
        # extension class 'ec': could we/should we use this as an
        # additional check for extension classes?

        # Breadth-first traversal of class1's superclass tree.  Order
        # doesn't matter because we're just looking for a "yes/no"
        # answer from the tree; if we were trying to resolve a name,
        # order would be important!
        stack = [class1]
        while stack:
            if stack[0] is class2:
                return 1
            stack.extend(list(stack[0].__bases__))
            del stack[0]
        else:
            return 0

    # Not a regular class, not an extension class: blow up for consistency
    # with builtin 'issubclass()"
    else:
        raise TypeError, "arguments must be class or ExtensionClass objects"

# issubclass ()

def isinstance (object, klass):
    """A version of 'isinstance()' that works with extension classes
    as well as regular Python classes."""

    if type(klass) is TypeType:
        return __builtin__.isinstance(object, klass)
    elif hasattr(object, '__class__'):
        return issubclass(object.__class__, klass)
    else:
        return 0


import types

# This is the list of types always defined in Python 2.0's types.py.
# Some types (complex, unicode) are optional at build-time; others were
# added in Python 2.2.  Those are handled separately so Grouch will work
# with the lowest-common-denominator Python 2.0.

_type_name = {
    # everyday types
    types.DictType             : 'dict',
    types.FileType             : 'file',
    types.FloatType            : 'float',
    types.IntType              : 'int',
    types.ListType             : 'list',
    types.LongType             : 'long',
    types.NoneType             : 'none',
    types.StringType           : 'string',
    types.TupleType            : 'tuple',

    # callable things
    types.BuiltinFunctionType  : 'builtin function',
    types.FunctionType         : 'function',
    types.MethodType           : 'instance method',

    # not-quite-everyday types
    types.ClassType            : 'class',
    types.InstanceType         : 'instance',
    types.ModuleType           : 'module',
    types.TypeType             : 'type',

    # esoterica
    types.CodeType             : 'code',
    types.FrameType            : 'frame',
    types.TracebackType        : 'traceback',
    types.BufferType           : 'buffer',
    types.EllipsisType         : 'ellipsis',
    types.SliceType            : 'slice',
    types.XRangeType           : 'xrange',
}

# These are optional in Python 2.0, or were added in 2.2.
for (name, typename) in [("ComplexType", "complex"),
                         ("UnicodeType", "unicode"),
                         ("DictProxyType", "dict-proxy"),
                         ("GeneratorType", "generator"),
                         ("ObjectType", "object")]:
    if hasattr(types, name):
        _type_name[getattr(types, name)] = typename

def get_type_name (type):
    global _type_name
    return _type_name.get(type, type.__name__)



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/schema_util.py $
$Id: schema_util.py 24750 2004-07-21 15:26:51Z dbinger $

Search a set of directories for Python modules and classes, and parse
the class docstrings to generate an object schema.  Lots of
project-specific information (eg. directories to search, classes to
exclude, ...) can be supplied via a project description file.
"""

import sys, os, string, re
import getopt
import types
import traceback
from fnmatch import fnmatch
from cPickle import dump

from grouch.schema import ObjectSchema, ClassDefinition, InvalidAlias
from grouch.script_util import announce, error, warn


# -- Mid-level workers -------------------------------------------------
# (called by generate_schema())

def find_modules (dirs, base_dir=None, prefix=None, exclude=None):
    """find_modules(dirs : [string],
                    base_dir : string = None,
                    prefix : string = None,
                    exclude : [string] = None)
       -> [(modname : string, filename : string)]

    Searches a list of directories for Python module files (*.py).
    Returns a list of (modname, filename) tuples.  If 'prefix' is
    supplied, it is prepended (with a dot interpolated) to each module
    name.  If 'exclude' is supplied, it must be a list of
    fully-qualified module names; if a module is found in 'exclude'
    (after adding prefix), it will not be included in the returned list
    of modules.

    Example: dirs == ["foo"] and directory "foo" contains "bar.py" and
    "baz.py".  If no prefix or base_dir is supplied, returns
      [("foo.bar", "foo/bar.py"),
       ("foo.baz", "foo/baz.py")]

    If base_dir == "d" and prefix == "p", returns
      [("p.foo.bar", "d/foo/bar.py"),
       ("p.foo.baz", "d/foo/baz.py")]
    """
    modules = []                        # list of (modname, filename) tuples
    for dir in dirs:
        components = os.path.normpath(dir).split(os.sep)
        if os.pardir in components:     # disallow "../foo" after normpath()
            raise ValueError, \
                  ("invalid directory '%s': cannot contain '%s'" %
                   (dir, os.pardir))
        if components == [os.curdir]: # eg. dir == "." or ""
            components = []
        if prefix:
            components.insert(0, prefix)

        if base_dir:
            real_dir = os.path.normpath(os.path.join(base_dir, dir))
        else:
            real_dir = os.path.normpath(dir)

        for basename in os.listdir(real_dir):
            # Skip files that aren't Python modules
            if not fnmatch(basename, "*.py") or basename == "__init__.py":
                continue

            real_filename = os.path.join(real_dir, basename)
            bare_modname = os.path.splitext(basename)[0]
            modname = ".".join(components + [bare_modname])

            if modname not in exclude:
                modules.append((modname, real_filename))

    return modules


class ClassInfo:
    """
    Encapsulates all the information we need about a class in order to
    add it to an object schema.  find_classes() returns a list of
    ClassInfo instances to find_all_classes(), which incorporates that
    in a dictionary mapping module names to lists of ClassInfo
    instances.

    Instance attributes:
      bare_name : string
        the bare class name as seen in the "class" statement
      full_name :
        the fully-qualified class name (ie. including module name)
      base_classes : [string]
        the list of base class names from the class statement
      docstring : string
        the class docstring as a single (probably multi-line) string
    """

    def __init__ (self, bare_name, full_name, base_classes, docstring):
        self.bare_name = bare_name
        self.full_name = full_name
        self.base_classes = base_classes
        self.docstring = docstring

    def __str__ (self):
        return self.full_name

    def __repr__ (self):
        return "<%s at %08x: %s>" % (self.__class__.__name__, id(self), self)

    def expand_base_classes (self, schema):
        """
        Ensure that every base class listed in self.base_classes is
        either a class or an alias for a class in 'schema'.  Any base
        class names that are aliases are expanded to the full class
        name.  Raises ValueError if any base class names are bad.
        """
        for i in range(len(self.base_classes)):
            base_name = self.base_classes[i]
            if not schema.get_class_definition(base_name):
                alias = schema.get_alias(base_name)
                if not alias:
                    raise ValueError(
                        "%s: invalid base class %r (no such class or alias)"
                        % (self.full_name, base_name))
                elif not alias.is_plain_instance_type():
                    raise ValueError(
                        "%s: invalid base class %r (alias to non-class)"
                        % (self.full_name, base_name))
                else:
                    self.base_classes[i] = alias.klass_name


def find_all_classes (modules, schema, exclude_classes=None):
    """find_all_classes(modules : [(modname : string, filename : string)],
                        schema : ObjectSchema,
                        exclude_classes : [string] = None)
       -> { modname:string : [ClassInfo] }

    Find all classes in a list of modules.  Returns a dictionary mapping
    module name to list of ClassInfo objects.  For each class found,
    adds two things to schema: an empty ClassDefinition, and an alias
    mapping the bare class name to its full name (eg. for class
    foo.bar.FooBar, the alias maps "FooBar" to "foo.bar.FooBar").
    """

    # Map module name to list of class objects
    module_classes = {}
    num_classes = 0

    announce("looking for classes...\n")
    for (modname, filename) in modules:
        announce("module %s:\n" % modname, threshold=2)
        klasses = find_classes(filename, modname, exclude_classes)
        num_classes += len(klasses)

        ok_klasses = []
        for klass in klasses:
            bare_name = klass.bare_name
            full_name = klass.full_name

            # create new alias mapping bare_name to full_name and add
            # it to the schema?
            add_alias = 1

            alias = schema.get_alias(bare_name)
            if isinstance(alias, InvalidAlias):
                # We've already had one name collision on 'bare_name' and
                # invalidated that alias, so don't try to create it again!
                warn("can't alias %r to %r: name collision"
                     % (bare_name, full_name))
                add_alias = 0

            elif alias:
                # The alias is actually for this class, so it's not a
                # problem after all.
                if (alias.is_instance_type() and
                    alias.get_class_name() == full_name):
                    add_alias = 0

                # Uh-oh, alias already exists and it maps 'bare_name'
                # to something else.  Ambiguity is bad, so invalidate
                # this alias.
                else:
                    warn("invalidating ambiguous alias %r"
                         % bare_name)
                    schema.invalidate_alias(bare_name)
                    add_alias = 0

            # If we need to, add a new class definition to the schema.
            klass_def = schema.get_class_definition(full_name)
            if klass_def is None:
                klass_def = ClassDefinition(full_name, schema)
                schema.add_class(klass_def)

            # The above code decided there's no problem aliasing
            # 'bare_name' to 'full_name', so do it now.
            if add_alias:
                schema.add_alias(bare_name, full_name)

        module_classes[modname] = klasses
        for klass in klasses:
            announce("  %s\n" % klass, threshold=2)

    # This has to be done after scanning all modules, because it depends
    # on us having seen all classes in the application.
    for klasses in module_classes.values():
        for klass in klasses:
            try:
                klass.expand_base_classes(schema)
            except ValueError, err:
                # Should we exclude the class from the schema entirely
                # if this happens?  Right now we just refrain from giving
                # it any base classes if one of its base classes has
                # a problem.
                error(str(err))
            else:
                klass_def = schema.get_class_definition(klass.full_name)
                klass_def.set_bases(klass.base_classes)

    announce("found %d classes\n" % num_classes)
    return module_classes


def get_names (nodes):
    from compiler import ast
    names = []
    for node in nodes:
        if isinstance(node, ast.Name):
            names.append(node.name)
        elif isinstance(node, ast.Getattr):
            # Dotted name: "foo.bar.baz" in source becomes
            # Getattr(Getattr(Name(foo),bar),baz) in the AST.  Unwind
            # this stack to reconstruct the original string.
            cur_node = node
            stack = []
            while isinstance(cur_node, ast.Getattr):
                stack.append(cur_node.attrname)
                cur_node = cur_node.expr
            assert isinstance(cur_node, ast.Name), \
                   "expected Name at bottom of Getattr stack"
            stack.append(cur_node.name)
            stack.reverse()
            name = string.join(stack, ".")
            names.append(name)

    return names


def find_classes (filename, modname, exclude=None):
    """find_classes(filename : string,
                    modname : string,
                    exclude : [string] = None)
       -> [ClassInfo]

    Parses the Python source file 'filename' looking for classes.
    Assumes this source file contains the module 'modname'.  Returns a
    list of ClassInfo instances, one for each class in the file.

    Eg. if parsing a file foo.py which (according to 'modname') contains
    a module 'foo', this code:
      class Foo (bar.Bar):
        '''
        foo
        bar
        '''
    results in a ClassInfo instance like:
      bare_name = "Foo"
      full_name = "foo.Foo"
      base_classes = ["bar.Bar"]
      docstring = "\n    foo\n    bar\n    "

    No attempt is made to determine what "bar.Bar" really refers to --
    that would require interpreting the module, and in that case we
    might as well just import the damn thing.  Note that the
    ClassInfo.expand_base_classes() method, called by find_all_classes()
    just before returning, attempts to expand all base class names to
    their true names.
    """
    from compiler import parseFile, ast
    from parser import ParserError

    klasses = []

    try:
        module_ast = parseFile(filename)
    except ParserError:
        error("%s: unable to parse module (try importing it for more details)"
              % filename)

    for node in module_ast.node.nodes:
        if isinstance(node, ast.Class):
            if modname:
                fullname = modname + "." + node.name
            else:
                fullname = node.name

            if fullname in exclude:
                continue

            base_classes = get_names(node.bases)
            klasses.append(ClassInfo(node.name, fullname,
                                     base_classes, node.doc))

    return klasses


def parse_class_docstrings (modules, module_classes, schema):
    """parse_class_docstrings(modules : [(string, string)],
                              module_classes : {string : [ClassInfo]},
                              schema : ObjectSchema)
    """
    announce("parsing class docstrings...\n")
    for (modname, _) in modules:
        klasses = module_classes.get(modname, [])
        announce("  module: %s (%d classes)\n" % (modname, len(klasses)),
                 threshold=2)
        for klass in klasses:
            try:
                errors = parse_docstring(klass, schema)
            except ValueError, exc:
                error(str(exc))
                announce("   failed to parse %s docstring\n" % klass, threshold=2)
            else:
                for e in errors:
                    warn(e)
                announce("   parsed %s docstring\n" % klass, threshold=2)


# -- Parsing code ------------------------------------------------------
# (parse_class_docstrings() calls parse_docstring(), which uses
#  everything else in this section)

leading_ws_re = re.compile('^(\s*)')

def get_indent_level (s):
    """Return the number of spaces that 's' starts with."""
    m = leading_ws_re.match(s)
    return len(m.group(1))


def clean_docstring (doc, klass_name):
    lines = (doc or '').split("\n")
    assert lines, "string.split() returned empty list"
    if lines[0] and lines[0][0] == ' ':              # first line indented
        leading_indent = get_indent_level(lines[0])
        start_line = 0
    elif len(lines) > 1:                # look for first "real" line
        i = 1
        while i < len(lines) and not lines[i]: # skip over blank lines
            i += 1
        assert i < len(lines), "arg, screwy docstring"
        leading_indent = get_indent_level(lines[i])
        start_line = 1
    else:                               # single-line docstring
        leading_indent = 0
        start_line = 0

    for i in range(start_line, len(lines)):
        if not lines[i]:                # skip blanks
            continue
        if len(lines[i]) < leading_indent:
            raise ValueError, \
                  ("class %s: inconsistent indent in docstring "
                   "(line %d too short)" % (klass_name, i))
        if string.lstrip(lines[i][:leading_indent]) != "":
            raise ValueError, \
                  ("class %s: inconsistent indent in docstring "
                   "(line %d dedented relative to line %d)" %
                   (klass_name, i, start_line))

        lines[i] = re.sub(r'#.*', '', lines[i])
        lines[i] = string.rstrip(lines[i][leading_indent:])

    return lines

def find_attrs (lines, klass_name):
    i = 0
    while i < len(lines):
        line = lines[i]
        if line.startswith("Instance attributes:"):
            if line.endswith("none"):
                return None
            return i+1
        i += 1
    else:
        raise ValueError, \
              ("class %s: no \"Instance attributes:\" line in docstring" %
               klass_name)


_name_pat = r'[a-zA-Z_][a-zA-Z0-9_]*'
_dotted_name_pat = r'%s(?:\.%s)*' % (_name_pat, _name_pat)
_attr_line_re = re.compile(r'\s*(%s)\s*:\s*(.*)' % _name_pat)
_element_name_re = re.compile(r'(%s):(%s)' % (_name_pat, _dotted_name_pat))

def massage_typespec (typespec):
    # Now we have to massage the typespec so it can be parsed as
    # a ValueType.  Luckily, the docstring type specification
    # language is pretty similar to the ValueType type
    # specification language; the differences are:
    #   - container elements can have names as well as types,
    #     eg.  { key:keytype : value:valuetype }
    #     or   (val1:type1, val2:type2)
    #   - plain container types can include the name of the
    #     container, eg.
    #       dictionary {string : int}
    #   - typespecs can be trailed by default value description, eg.
    #       foo : int = 37
    #       bar : [string] = ["hello"]

    # Note that for named container elements, whitespace
    # matters!  {key:keytype : value:valuetype} is *not* the
    # same as {key : keytype : value : valuetype} -- the latter
    # is illegal.

    # Deal with the second exception first, since it's easiest.
    words = string.split(typespec, None, 1)
    if len(words) > 1:
        remainder = words[1]
        if words[0] in ('list', 'tuple', 'dictionary'):
            typespec = remainder
    else:
        remainder = None

    # Strip anything that looks like a default value, ie. " = ..."
    typespec = re.sub(r'\s*=.*', '', typespec)

    # Deal with the named-elements thing.  This is a kludge, but it
    # should work given the tight syntax constraints on named elements.
    if typespec[0] in "[({" or (remainder and remainder[0] in "[({"):
        typespec = _element_name_re.sub(r'\2', typespec)

    return typespec


def parse_docstring(klass, schema):
    """(klass : ClassInfo, schema : ObjectSchema) -> errors : [string]

    Parse the class docstring in 'klass' and use the docstring to update
    the class definition already in 'schema'.  Return a list of error
    messages resulting from parsing the docstring which should be presented
    to the user.  Raises ValueError if the docstring is a lost cause,
    ie. missing or completely unparseable.
    """

    klass_name = klass.full_name        # for error messages

    klass_def = schema.get_class_definition(klass_name)
    if klass_def.attrs or klass_def.all_attrs:
        raise ValueError, "class %s: already seen" % klass_name

    doc = klass.docstring
    if not doc:
        raise ValueError, "class %s: no docstring" % klass_name

    errors, attrs = parse_docstring_helper(schema, doc, klass_name)

    for name, type in attrs:
        if type.is_any_type():
            type.set_allow_any_instance(0)
        klass_def.add_attribute(name, type)

    # XXX ClassDefinition should expose len() of its attrs list
    if len(klass_def.attrs) == 0:
        errors.append("class %s: no attributes successfully parsed" %
                      klass_def.name)

    return errors


def parse_docstring_helper(schema, doc, klass_name):
    """
    This is separated so that it can be used by outside code.
    """

    errors = []
    attrs = []

    lines = clean_docstring(doc, klass_name)
    # Now 'lines' is a list of consistently-indented lines.  Find the
    # one that says "Instance attributes:", and start parsing attribute
    # type specifications from there.  Stop when we get back to
    # zero-indent level.
    start_line = find_attrs(lines, klass_name)
    if start_line is None:              # class is declared to have no
        return errors, attrs            # instance attributes

    i = start_line
    while i < len(lines):
        line = lines[i]
        if not line:                    # skip blanks
            i += 1
            continue
        indent = get_indent_level(line)
        if indent == 0:                 # out of the attribute list
            break

        m = _attr_line_re.match(line)
        if m:
            (name, typespec) = m.group(1,2)
            typespec = massage_typespec(typespec)
            try:
                attrs.append((name, schema.parse_type(typespec)))
            except ValueError, exc:
                errors.append("class %s, attribute %s: %s" %
                              (klass_name, name, exc))
        else:
            errors.append("class %s: couldn't parse line %d of docstring: %s" %
                          (klass_name, i, `line`))

        # Read lines until back at the indent level of this line
        # (ie. skip the indented attribute description).
        i += 1
        while i < len(lines):
            if get_indent_level(lines[i]) <= indent:
                break
            i += 1
    return errors, attrs


# -- High-level workers ------------------------------------------------
# (called from main())

def generate_schema (project, base_dir):
    schema = ObjectSchema()

    # Preparatory work -- stuff that needs to go in the schema, but
    # can't be discovered by searching for and parsing *.py files.
    for atomic_type in project.atomic_types:
        if type(atomic_type) is types.TupleType and len(atomic_type) == 2:
            schema.add_atomic_type(*atomic_type)
        else:
            schema.add_atomic_type(atomic_type)

    for name in project.forward_classes:
        cdef = ClassDefinition(name, schema)
        schema.add_class(cdef)

    for (name, value) in project.type_aliases:
        schema.add_alias(name, value)

    # The meat of the schema: class definitions for classes in whatever
    # *.py files we can find in project.dirs.
    if project.dirs:
        announce("searching for modules...")
        modules = find_modules(project.dirs,
                               base_dir=base_dir,
                               prefix=project.prefix,
                               exclude=project.exclude_modules)
        announce("found %d modules\n" % len(modules))
    else:
        modules = []

    project.add_extra_modules(modules, base_dir)

    module_classes = find_all_classes(modules, schema, project.exclude_classes)
    parse_class_docstrings(modules, module_classes, schema)

    if project.post_parse_hook:
        project.post_parse_hook(schema)

    # Finish up all class definitions -- ie. look at the inheritance tree
    # and gather up the list of all attributes that should be in instances
    # of a class, including those inherited from superclasses.
    for klass_name in schema.get_class_names():
        klass_def = schema.get_class_definition(klass_name)
        klass_def.finish_definition()

    return schema


def write_schema (schema, text_filename, pickle_filename):

    if text_filename:
        announce("writing object schema to %s..." % text_filename)
        schema_file = open(text_filename, "w")
        schema.write_aliases(schema_file)
        schema_file.write("\n\n")

        for klass_name in schema.get_class_names():
            klass_def = schema.get_class_definition(klass_name)
            klass_def.write(schema_file)
            schema_file.write("\n")
        schema_file.close()
        announce("\n")

    if pickle_filename:
        announce("pickling object schema to %s..." % pickle_filename)
        schema_file = open(pickle_filename, "w")
        dump(schema, schema_file, 1)
        schema_file.close()
        announce("\n")


# -- Main program ------------------------------------------------------

class ProjectDescription:
    """
    Instance attributes:
      atomic_types : [(any, string) | any]
      forward_classes : [string]
      type_aliases : [(alias:string, alias_expansion:string)]
      prefix : string
      dirs : [string]
      extra_modules : [string | (string, string)]
      exclude_modules : [string]
      exclude_classes : [string]

      post_parse_hook : function
    """

    def __init__ (self):
        self.atomic_types = []
        self.forward_classes = []
        self.type_aliases = []
        self.prefix = None
        self.dirs = []
        self.extra_modules = []
        self.exclude_modules = []
        self.exclude_classes = []

        self.post_parse_hook = None

    def read (self, filename):
        data = {}
        if not os.path.isfile(filename):
            raise ValueError, "no such file: %s" % filename
        try:
            execfile(filename, data)
        except:
            (t, v, tb) = sys.exc_info()
            error("error in %s:" % filename)
            exc = "".join(traceback.format_exception_only(t, v))
            sys.stderr.write(exc)
            sys.exit(1)

        for (name, value) in data.items():
            if hasattr(self, name):
                setattr(self, name, value)

        self.check_types(filename)

    def check_types (self, filename):
        from grouch.context import TypecheckContext

        schema = ObjectSchema()
        schema.add_atomic_type(lambda: None) # add 'function' type
        cdef = ClassDefinition(self.__class__.__name__, schema)
        schema.add_class(cdef)
        klass = ClassInfo("ProjectDescription",
                          "ProjectDescription",
                          [], self.__class__.__doc__)
        errors = parse_docstring(klass, schema)
        assert not errors, "errors in my own docstring!"
        cdef.finish_definition()
        context = TypecheckContext(report_errors=0)
        schema.check_value(self, context)
        if context.num_errors() > 0:
            error("type errors in %s:" % filename)
            context.write_errors(sys.stderr)
            sys.exit(1)

    def add_extra_modules (self, modules, base_dir):
        """add_extra_modules(modules : [(string, string)])

        Adds the modules listed in self.extra_modules to modules.  A bit
        tricky because extra_modules might not include filenames, in
        which case we have to go out and find the files.
        """

        # extra_modules is a list of any of the following
        #   string
        #     fully-qualified module name -- we'll have to search
        #     sys.path to find the file
        #   (modname : string, filename : string)
        #     module name with filename; we have to make sure the
        #     file exists, and try it relative to base_dir if not

        # yow! this is hairy and undertested!

        for module in self.extra_modules:
            if type(module) is types.StringType:
                modname = module
                comps = module.split(".")
                tail = os.path.join(*comps) + ".py"
                for dir in sys.path:
                    filename = os.path.join(dir, tail)
                    if os.path.exists(filename):
                        break
                else:
                    raise ValueError(
                        "no such module %s: %s not found in sys.path"
                        % (module, tail))

            elif type(module) is types.TupleType and len(module) == 2:
                (modname, filename) = module
                if os.path.exists(filename):
                    pass                # found it, good
                elif base_dir:          # we have a second chance
                    filename2 = os.path.join(base_dir, filename)
                    if os.path.exists(filename2):
                        filename = filename2
                    else:
                        raise ValueError(
                            "no such module %s: neither %s nor %s exist"
                            % (modname, filename, filename2))
                else:
                    raise ValueError(
                        "no such module %s: %s does not exist"
                        % (modname, filename))
            else:
                raise TypeError("bad extra_modules")

            assert os.path.exists(filename)
            modules.append((modname, filename))

        # for module

    # add_extra_modules ()


def main ():
    global VERBOSITY                    # global because announce(

    prog = os.path.basename(sys.argv[0])
    args = sys.argv[1:]

    usage = """\
usage: %s [options]
options:
  -v                        verbose: run noisily (repeat for more noise)
  -q                        run quietly
  -p FILE, --project=FILE   read project description info from FILE
  -o FILE, --output=FILE    output file: write pickled schema to FILE
                            [default: schema.pkl]
  -t FILE, --text=FILE      text output: write human-readable schema to FILE
                            [default: none]
  -d DIR, --base-dir=DIR    interpret DIRS in project description file
                            relative to DIR
""" % prog

    try:
        (opts, args) = getopt.getopt(args, "p:o:t:d:vq",
                                     ["project=",
                                      "output=",
                                      "text=",
                                      "base-dir=",])
    except getopt.error, msg:
        sys.exit(usage + str(msg))

    project = ProjectDescription()
    VERBOSITY = 1
    text_filename = None
    pickle_filename = "schema.pkl"
    base_dir = None
    prefix = None
    for (opt, val) in opts:
        if opt in ("-p", "project"):
            project.read(val)
        elif opt in ("-o", "--output"):
            pickle_filename = val
        elif opt in ("-t", "text"):
            text_filename = val
        elif opt in ("-d", "--base-dir"):
            base_dir = val
        elif opt == "-v":
            VERBOSITY += 1
        elif opt == "-q":
            VERBOSITY = 0

    if len(args) != 0:
        raise SystemExit, usage + "error: too many arguments"

    # Find modules in project.dirs, classes in those modules, and parse
    # the docstring of each class.  The result is an ObjectSchema, a
    # collection of atomic types, type aliases, and class definitions.
    schema = generate_schema(project, base_dir)

    # Write the schema out to either or both of a text and pickle file.
    write_schema(schema, text_filename, pickle_filename)

# main ()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit("interrupted")



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/script_util.py $
$Id: script_util.py 24750 2004-07-21 15:26:51Z dbinger $
"""

import sys

def announce (msg, threshold=1):
    import __main__
    if __main__.VERBOSITY >= threshold:
        sys.stdout.write(msg)
        sys.stdout.flush()

def error (msg):
    sys.stderr.write("error: " + msg + "\n")

def warn (msg):
    sys.stderr.write("warning: " + msg + "\n")



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/valuetype.py $
$Id: valuetype.py 24750 2004-07-21 15:26:51Z dbinger $

Provides the ValueType class and its subclasses, which together
represent the Python type system as interpreted by Grouch.
"""

import string
import types
from sets import Set
from copy import copy

from grouch.util import \
     isinstance, is_instance_object, \
     get_class_object, \
     get_full_classname, get_type_name


# XXX Type-checking issues yet to be resolved
#
#  * if an "any" type has 'allow_none' false, does it allow None?
#    eg. if 'a' is declared "a : any notnone", can 'a' be None?
#  * ditto for boolean types:
#    if 'f' is decalred "f : boolean notnone", can 'f' be None?
#  * what does an empty/uninitialized union type mean?  no values at
#    all are valid?  perhaps only None should be allowed?


class ValueType:
    """
    Instance attributes:
      schema : ObjectSchema
        the object schema (atomic types, aliases, and class definitions)
        relative to which we should interpret this type
      allow_none : boolean = true
        are values allowed to be None?
    """

    # class attribute 'metatype' must be defined by each subclass
    metatype = None


    def __init__ (self, schema):
        self.schema = schema
        self.allow_none = 1

        assert self.__class__ is not ValueType,\
               "ValueType is an abstract class"
        assert self.metatype is not None,\
               "metatype must be set by subclass"


    # Subclasses must provide '__str__()'

    def __repr__ (self):
        return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self)


    # -- Accessors/modifiers -------------------------------------------

    def get_metatype (self):
        return self.metatype            # really a class attribute!


    # -- Predicates ----------------------------------------------------
    # XXX might be cleaner (and more extensible) to just "return 0"
    # everywhere here, and require subclasses to override and
    # "return 1" where applicable

    def is_atomic_type (self):
        return self.metatype == "atomic"

    def is_instance_type (self):
        return self.metatype in ("instance", "instance-container")

    def is_plain_instance_type (self):
        return self.metatype == "instance"

    def is_container_type (self):
        return self.metatype in ("container", "instance-container")

    def is_plain_container_type (self):
        return self.metatype == "container"

    def is_instance_container_type (self):
        return self.metatype == "instance-container"

    def is_union_type (self):
        return self.metatype == "union"

    def is_any_type (self):
        return self.metatype == "any"

    def is_boolean_type (self):
        return self.metatype == "boolean"

    def is_alias_type (self):
        return self.metatype == "alias"


    # -- Type-checking -------------------------------------------------

    def simple_check (self, value):
        raise NotImplementedError, "subclasses must implement"

    def check_value (self, value, context):
        raise NotImplementedError, "subclasses must implement"

    def check_none_ok (self, value):
        return (value is None and self.allow_none)

    def check_unexpected_instance (self, value, context, allow_foreign=0):
        if not is_instance_object(value):
            return 1
        try:
            itype = self.schema.make_instance_type(value.__class__)
        except ValueError, msg:
            # XXX ugh, should probably have some custom exceptions
            msg = str(msg)              # just in case
            if not msg.startswith("class not in schema"):
                raise

            if allow_foreign:
                return 1
            else:
                context.add_error("found instance of class not in schema: %s" %
                                  get_full_classname(value.__class__))
                return 0
        else:
            return itype.check_value(value, context)

    def count_object (self, context):
        context.count_object(self)

    def uncount_object (self, context):
        context.uncount_object(self)


class AtomicType (ValueType):
    """
    Instance attributes:
      type : type
        the type object that values must match
      type_name : string
        the name of the atomic type
    """

    metatype = "atomic"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.type = None
        self.type_name = None

    def __str__ (self):
        return self.type_name or "*no type*"

    # -- Pickle support ------------------------------------------------
    # (needed because you can't pickle type objects)

    # Pickling AtomicType instances is a bit tricky, because each
    # AtomicType carries around a type object in its 'type' attribute,
    # but you can't pickle type objects.  Thus we have to get rid of the
    # type objects at pickle-time, but we can't reconstitute them at
    # unpickle-time, because we need the enclosing ObjectSchema to
    # lookup the type name from the type object.  Depending on what
    # order things were pickled in, our schema may or may not be
    # available or complete, so we can't depend on it being there.
    #
    # The solution is to leave the 'type' attribute None (which is what
    # it's set to at pickle-time), and then lookup the type object in
    # the schema only when we need it, ie. when either 'get_type()' or
    # 'check_value()' is called.

    def __getstate__ (self):
        if self.type is None:
            return self.__dict__
        else:
            dict = copy(self.__dict__)
            dict['type'] = None
            return dict


    # -- Accessors/modifiers -------------------------------------------

    def set_type (self, tipe):
        tipe = self.schema.get_atomic_type_object(tipe)
        self.type = tipe
        self.type_name = self.schema.get_atomic_type_name(tipe)

    def get_type (self):
        if self.type is None:
            self.type = self.schema.get_atomic_type_object(self.type_name)
        return self.type


    # -- Type-checking methods -----------------------------------------

    def simple_check (self, value):
        vtype = type(value)
        if self.type is None:           # first use after unpickling?
            self.type = self.schema.get_atomic_type_object(self.type_name)
        return (vtype is self.type) or self.check_none_ok(value)

    def check_value (self, value, context):
        self.count_object(context)
        if self.simple_check(value):
            return 1
        else:
            context.add_error("expected %s, got %s" %
                              (str(self), describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

    def count_object (self, context):
        context.count_object(self, self.type_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.type_name)

# class AtomicType


# XXX name collision w/ standard 'types' modules!
class InstanceType (ValueType):
    """
    Instance attributes:
      klass_name : string
        the (fully-qualified) name of the class
      klass_definition : ClassDefinition
        the definition of this class in the object schema (defines
        what attributes are allowed/expected and what their
        types are)
      klass : class
        a class object; values must be instances of this class

    XXX klass and klass_definition might be obsolete; they are now
    assigned only when needed, ie. in check_value()
    """

    metatype = "instance"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.klass_name = None
        self.klass_definition = None
        self.klass = None

    def __str__ (self):
        return self.klass_name or "*no class*"

    def set_class_name (self, klass_name):
        if type(klass_name) is not types.StringType:
            raise TypeError, "not a string: %r" % klass_name
        self.klass_name = klass_name

    def get_class_name (self):
        return self.klass_name


    # Helper methods for 'check_value()'

    def _find_classdef (self, value, context):
        klass = value.__class__
        if klass is self.klass:
            klass_def = self.klass_definition
        else:
            klass_def = self.schema.get_class_definition(klass)
            if not klass_def:
                context.add_error("expected instance of %s, "
                                  "got instance of unknown subclass %s"
                                  % (self.klass,
                                     get_full_classname(klass)))

        return klass_def

    def _check_attrs (self, value, klass_def, context):

        #print "checking " + repr(value)

        # First, make sure there are no unexpected attributes.  (Doing this
        # first -- in the instance dictionary's hash order -- means the
        # *real* type-checking loop, below, is done according to the order
        # of attributes in the class definition, ie. in a predictable,
        # canonical order.  This is a good thing, if only because it makes
        # writing tests easier.)

        if context.mode == "zodb":
            # prod ZODB into loading the object before relying on
            # the value of dir()
            dir(value)
            #print "activated %r" % value

            if hasattr(value, '_p_oid'):
                obj_id = "oid %r, id %x" % (value._p_oid, id(value))
            else:
                obj_id = "id %x" % id(value)


        all_ok = 1

        # Some ExtensionClasses (OOBTree in particular) don't provide
        # a __dict__ for their instances; treat that as an empty
        # instance attribute dictionary.
        if hasattr(value, '__dict__'):
            actual_dict = vars(value)
        else:
            actual_dict = {}

        for (actual_attr, val) in actual_dict.items():
            if actual_attr.startswith('__') and actual_attr.endswith('__'):
                continue
            if context.mode == "zodb" and actual_attr.startswith("_v_"):
                continue

            if not klass_def.get_attribute_type(actual_attr):
                context.add_error("found unexpected attribute '%s': %s" %
                                  (actual_attr, describe_value(val)))

                context.push_context('attr', actual_attr)
                self.check_unexpected_instance(val, context)
                context.pop_context()

                all_ok = 0

        # Now the real type-checking loop, in canonical (not hash)
        # order.  By-the-by, we also check for missing attributes here.
        for expected_attr in klass_def.all_attrs:
            if actual_dict.has_key(expected_attr):
                val = actual_dict[expected_attr]
                expected_type = klass_def.get_attribute_type(expected_attr)
                context.push_context('attr', expected_attr)
                ok = expected_type.check_value(val, context)
                if not ok:
                    all_ok = 0

                context.pop_context()
            else:
                context.add_error("expected attribute '%s' not found" %
                                  expected_attr)
                all_ok = 0

        if context.mode == "zodb" and hasattr(value, '_p_deactivate'):
            #print "deactivating %r" % value
            value._p_deactivate()

        return all_ok

    def simple_check (self, value):
        return (self.check_none_ok(value) or
                isinstance(value, get_class_object(self.klass_name)))

    def check_value (self, value, context):

        # XXX from Jim Fulton on zodb-dev@zope.org, about how to
        # reduce memory use of the type-checking script:
        #
        # You could modify your script to your script to "ghostify" the
        # objects, by calling _p_deactivate on them after you use them
        # to prevent the memory bloat.  This is what the Zope "find"
        # facility does when searching for objects.
        #
        # I think this is the place to do it, but this method would need
        # a bit of rearrangement.

        if self.klass is None:
            self.klass = get_class_object(self.klass_name)
        if self.klass_definition is None:
            self.klass_definition = \
                self.schema.get_class_definition(self.klass_name)

        self.count_object(context)
        if isinstance(value, self.klass):
            klass_def = self._find_classdef(value, context)
            if not klass_def:
                return 0

            if context.check_seen(value):
                # No new errors to add here, so act as though the guts of
                # the object are fine.
                self.uncount_object(context)
                return 1

                # XXX maybe the context should keep track of which
                # instances/containers had errors in them, so we can say
                # "second visit to object with errors" -- hmm, that'll only
                # work with instances, because a container might visited
                # twice with different declarations in force.

            #print "checking", repr(value)
            return self._check_attrs(value, klass_def, context)

        elif self.check_none_ok(value):
            return 1

        else:
            context.add_error("expected instance of %s, got %s" %
                              (self.klass, describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

        # if/elif/else

    # check_value ()

    def count_object (self, context):
        context.count_object(self, self.klass_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.klass_name)

# class InstanceType


class ContainerType (ValueType):
    """
    Instance attributes:
      allow_empty : boolean = true
        only for container and instance-container types: are
        values allowed to be empty containers?
    """

    metatype = "container"
    container_type = None               # subclasses must set

    def __init__ (self, schema):
        assert self.__class__ is not ContainerType,\
               "ContainerType is an abstract class"

        ValueType.__init__(self, schema)
        self.allow_empty = 1


    # -- Type-checking methods -----------------------------------------

    container_type_object = {
        "list" : types.ListType,
        "tuple" : types.TupleType,
        "dictionary" : types.DictionaryType,
        "set" : Set,
        }

    def simple_check (self, value):
        return (self.check_none_ok(value) or
                type(value) is self.container_type_object[self.container_type])

    def check_value (self, value, context):
        self.count_object(context)
        if type(value) is self.container_type_object[self.container_type]:
            if len(value) == 0 and self.container_type != 'tuple':
                if not self.allow_empty:
                    context.add_error("empty %s not allowed" %
                                      self.container_type)
                    return 0
                else:
                    return 1

            return self.check_contents(value, context)
        elif self.check_none_ok(value):
            return 1
        else:
            context.add_error("expected %s, got %s" %
                              (self.container_type, describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

    def count_object (self, context):
        context.count_object(self, self.container_type)

    def uncount_object (self, context):
        context.uncount_object(self, self.container_type)

# class ContainerType


class ListType (ContainerType):
    """
    Instance attributes:
      element_type : ValueType
        the type that all elements of the list must be
    """

    container_type = "list"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.element_type = None

    def __str__ (self):
        return "[%s]" % self.element_type

    def set_element_type (self, element_type):
        self.element_type = self.schema.get_type(element_type)

    set_list_type = set_element_type    # backwards compat.

    def get_element_type (self):
        return self.element_type


    # -- Type-checking methods -----------------------------------------

    def check_contents (self, value, context):
        i = 0
        all_ok = 1
        for item in value:
            context.push_context('item', i)
            ok = self.element_type.check_value(item, context)
            if not ok:
                all_ok = 0
            context.pop_context()
            i = i + 1

        return all_ok


class TupleType (ContainerType):
    """
    Instance attributes:
      element_types : [ValueType]
        the type of each "slot" in the tuple (values of this type
        must be tuples of the same length as 'element_types', and
        each element of the tuple must match the corresponding
        type in 'element_types'
      extended : boolean = false
        if true, this type is for "extended" tuples
    """

    container_type = "tuple"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.element_types = []
        self.extended = 0

    def __str__ (self):
        types = self.element_types
        if self.extended:
            if len(types) == 1:
                return "(%s*,)" % types[0]
            else:
                types = map(str, types)
                return ("(%s, %s*)" %
                        (string.join(types[0:-1], ", "), types[-1]))

        else:                           # regular tuple type
            if len(types) == 1:
                return "(%s,)" % types[0]
            else:
                types = map(str, types)
                return "(%s)" % string.join(types, ", ")

    def set_element_types (self, element_types):
        if type(element_types) not in (types.TupleType, types.ListType):
            raise TypeError, \
                  "not a sequence of type objects: %s" % `element_types`
        self.element_types = map(self.schema.get_type, element_types)

    def set_extended (self, extended):
        self.extended = extended

    set_tuple_type = set_element_types  # backwards compat.

    def get_element_types (self):
        return self.element_types

    def is_extended (self):
        return self.extended

    def check_contents (self, value, context):
        types = self.element_types
        ext = self.extended

        # First, just check the length; if len(element_types) == N:
        #  - regular tuples must be of length N
        #  - extended tuples must be of length >= N-1
        #    (last item repeats 0 .. * times)

        if (not ext and len(value) != len(types)):
            context.add_error("expected tuple of length %d, "
                              "got one of length %d" %
                              (len(types), len(value)))
            all_ok = 0
        elif (ext and len(value) < len(types)-1):
            context.add_error("expected tuple of at least length %d, "
                              "got one of length %d" %
                              (len(types)-1, len(value)))
            all_ok = 0
        else:
            all_ok = 1

        # Length is OK, so now check the actual elements
        if ext:                         # check the whole tuple
            n = len(value)
        else:                           # check only what we have types for
            n = min(len(value), len(types))

        for i in range(n):
            context.push_context('item', i)
            if ext and i >= len(types):
                cur_type = types[-1]
            else:
                cur_type = types[i]

            if not cur_type.check_value(value[i], context):
                all_ok = 0
            context.pop_context()

        return all_ok


class DictionaryType (ContainerType):
    """
    Instance attributes:
      key_type : ValueType
      value_type : ValueType

    """

    container_type = "dictionary"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.key_type = None
        self.value_type = None

    def __str__ (self):
        return "{ %s : %s }" % (self.key_type, self.value_type)

    def set_key_type (self, key_type):
        self.key_type = self.schema.get_type(key_type)

    def set_value_type (self, value_type):
        self.value_type = self.schema.get_type(value_type)

    def set_item_types (self, key_type, value_type):
        self.key_type = self.schema.get_type(key_type)
        self.value_type = self.schema.get_type(value_type)

    set_dictionary_type = set_item_types # backwards compat.


    def get_key_type (self):
        return self.key_type

    def get_value_type (self):
        return self.value_type

    def get_item_types (self):
        return (self.key_type, self.value_type)


    def check_contents (self, value, context):
        all_ok = 1
        for key in value.keys():
            context.push_context('dkey', key)
            if not self.key_type.check_value(key, context):
                all_ok = 0

            context.replace_context('dval', key)
            val = value[key]
            if not self.value_type.check_value(val, context):
                all_ok = 0

            context.pop_context()

        return all_ok


class SetType (ContainerType):
    """
    Instance attributes:
      element_type : ValueType
        the type that all elements of the set must be
    """

    container_type = "set"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.element_type = None

    def __str__ (self):
        return "{%s}" % self.element_type

    def set_element_type (self, element_type):
        self.element_type = self.schema.get_type(element_type)

    def get_element_type (self):
        return self.element_type


    # -- Type-checking methods -----------------------------------------

    def check_contents (self, value, context):
        i = 0
        all_ok = 1
        for item in value:
            context.push_context('item', i) # XXX sets are unordered, fix?
            ok = self.element_type.check_value(item, context)
            if not ok:
                all_ok = 0
            context.pop_context()
            i = i + 1

        return all_ok


class InstanceContainerType (InstanceType):
    """
    Instance attributes:
      container_type : ContainerType
    """

    metatype = "instance-container"

    def __init__ (self, schema):
        InstanceType.__init__(self, schema)
        self.container_type = None

    def __str__ (self):
        return "%s %s" % (self.klass_name, self.container_type)

    def set_container_type (self, container_type):
        if not isinstance(container_type, ContainerType):
            raise TypeError, \
                  "not a ContainerType instance: %s" % `container_type`
        self.container_type = container_type

    def get_container_type (self):
        return self.container_type

    def get_element_type (self):
        if not isinstance(self.container_type, ListType):
            raise RuntimeError, "not a list type"
        return self.container_type.get_element_type()

    def get_element_types (self):
        if not isinstance(self.container_type, TupleType):
            raise RuntimeError, "not a tuple type"
        return self.container_type.get_element_types()

    def get_item_types (self):
        if not isinstance(self.container_type, DictionaryType):
            raise RuntimeError, "not a dictionary type"
        return self.container_type.get_item_types()


    def simple_check (self, value):
        return (self.check_none_ok(value) or
                (self.container_type.simple_check(value) and
                 InstanceType.simple_check(self, value)))

    def check_value (self, value, context):
        if self.check_none_ok(value):
            return 1
        cont_ok = self.container_type.check_contents(value, context)
        inst_ok = InstanceType.check_value(self, value, context)
        return inst_ok and cont_ok


class UnionType (ValueType):
    """
    Instance attributes:
      union_types : [ValueType]
        a set of types; values must match one of them
    """

    metatype = "union"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.union_types = None

    def __str__ (self):
        types = map(str, self.union_types)
        return string.join(types, " | ")

    def set_union_types (self, union_types):
        if type(union_types) is not types.ListType:
            raise TypeError, "not a list of type objects: %s" % `union_types`
        union_types = map(self.schema.get_type, union_types)
        self.union_types = union_types

    def get_union_types (self):
        return self.union_types

    def simple_check (self, value):
        # XXX using sub-type's simple_check() here is not perfect:
        # consider what happens with a declaration like
        #   x : [int] | [string]
        # ...as long as x is a list, either branch is OK, so we'll
        # assume x is a list of ints.  If it's really a list of strings,
        # we'll barf unfairly all over it.  But it's not enough to
        # just go down one level into the list; what if that declaration
        # were
        #   x : [(int,string)] | [(string,int)]
        # ?  How far down do you have to go to differentiate the
        # branches of a UnionType?  I suspect this is a lost cause,
        # which is why I'm punting and leaving it obviously broken
        # (for the "[int] | [string]" case).

        for subtype in self.union_types:
            if subtype.simple_check(value):
                return subtype
        return None

    def check_value (self, value, context):
        self.count_object(context)

        subtype = self.simple_check(value)
        if subtype:
            ok = subtype.check_value(value, context)
            subtype.uncount_object(context)
            return ok
        else:
            types = self.union_types
            if len(types) == 0:
                prefix = "no valid values for empty union type"
            else:
                # Ugh, can forget about I18N of this code...
                tnames = map(str, types)
                if len(types) == 1:
                    tlist = tnames[0]
                elif len(types) == 2:
                    tlist = "%s or %s" % tuple(tnames)
                else:
                    tlist = (string.join(tnames[0:-1], ", ") +
                             ", or " + tnames[-1])
                prefix = "expected " + tlist

            context.add_error(prefix + "; got %s" % describe_value(value))
            self.check_unexpected_instance(value, context)
            return 0

    # check_value ()

    def count_object (self, context):
        context.count_object(self, str(self))

    def uncount_object (self, context):
        context.uncount_object(self, str(self))

# class UnionType


class AnyType (ValueType):
    """
    Instance attributes: none
      allow_any_instance : boolean
        if true (the default), class instances of any class are allowed;
        otherwise, instances must be of classes known to the containing
        object schema.
    """

    metatype = "any"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.allow_any_instance = 1

    def __str__ (self):
        return "any"

    def set_allow_any_instance (self, allow_any_instance):
        self.allow_any_instance = allow_any_instance

    def simple_check (self, value):
        return value is not None or self.allow_none

    def check_value (self, value, context):
        self.count_object(context)
        if self.simple_check(value):
            return self.check_unexpected_instance(
                value, context, allow_foreign=self.allow_any_instance)
        else:
            context.add_error("expected anything except None")
            return 0

class BooleanType (ValueType):
    """
    Instance attributes: none
    """

    metatype = "boolean"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)

    def __str__ (self):
        return "boolean"

    def simple_check (self, value):
        return value in (0, 1, None)

    def check_value (self, value, context):
        self.count_object(context)
        if self.simple_check(value):
            return 1
        else:
            context.add_error("expected a boolean, got %s" %
                              describe_value(value))
            self.check_unexpected_instance(value, context)
            return 0

class AliasType (ValueType):
    """
    Instance attributes:
      alias_name : string
        only for alias types: the name of this type (the alias)
      alias_type : ValueType
        the aliased type (what 'alias_name' stands for)
    """

    metatype = "alias"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.alias_name = None
        self.alias_type = None

    def __str__ (self):
        return self.alias_name

    def set_alias_type (self, alias_name, alias_type):
        self.alias_name = alias_name
        self.alias_type = alias_type

    def get_alias_type (self):
        return self.alias_type

    def simple_check (self, value):
        return self.alias_type.simple_check(value)

    def check_value (self, value, context):
        self.count_object(context)
        return self.alias_type.check_value(value, context)

    def count_object (self, context):
        context.count_object(self, self.alias_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.alias_name)

# -- Utility functions -------------------------------------------------

def describe_value (value):
    if is_instance_object(value):
        # It's an instance of either a vanilla Python class or an
        # ExtensionClass.
        return "instance of %s" % get_full_classname(value.__class__)
    elif value is None:
        return "None"
    else:
        if (type(value) in (types.ListType,
                            types.TupleType,
                            types.DictionaryType)
            and len(value) > 10):
            str_value = "... elided ..."
        elif type(value) is types.StringType and len(value) > 100:
            str_value = value[0:20] + " ... " + value[-20:]
        else:
            str_value = repr(value)

        return "%s (%s)" % (get_type_name(type(value)), str_value)



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/__init__.py $
$Id: __init__.py 24750 2004-07-21 15:26:51Z dbinger $

An system for describing and enforcing a Python object schema.
"""

